
/*
* xmp-aac - a XMPlay plugin for AAC/MP4 playback
* Copyright (C) 2004-2005, Sebastian Andersson, MaresWEB, http://www.maresweb.net
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include "xmpin.h"
#include "neaacdec.h"
#include "mp4ff/mp4ff.h"
#include "aacinfo.h"
#include "resource.h"

static HINSTANCE inst;
//static BOOL isWinNT;
static XMPIN_FUNCS* xmpfuncs;

static BOOL dmix=FALSE; // downmatrix
static BOOL disvbr=FALSE; // display VBR

#define MESS(win,id,m,w,l) SendDlgItemMessage(win,id,m,(WPARAM)w,(LPARAM)l)

typedef struct
{
	int mode; // 0 = AAC, 1 = MP4

	XMPFILE f;
	NeAACDecHandle decoder;
	float* buf; // decoding buffer
	BYTE* fbuf; // file buffer
	DWORD buflen,bufp,fbuflen,brate,abrate;
	BYTE nch;
	DWORD sr;
	float length;
	BOOL eof;

	int seq_samples,seq_bytes;

	// *** AAC PART ***
	int frames;
	faadAACInfo info;
	DWORD initbytes; // initialized bytes
	DWORD* st;
	BOOL seekable;

	// *** MP4 PART ***
	mp4ff_callback_t mp4cb;
	mp4ff_t* mp4file;
	long mp4track;
	int tagc;
	DWORD sampleid,timescale;
	long ns; // number of samples
	char** tags; // iTunes/MP4 metadata
} AACSTREAM;

static AACSTREAM* aac=NULL;

BOOL CALLBACK DialogProc(HWND h,UINT m,WPARAM w,LPARAM l)
{
	switch(m) {
		case WM_COMMAND:
			switch(LOWORD(w)) {
				case IDCANCEL:
					EndDialog(h,l);
					break;
				case IDOK:
					dmix=MESS(h,IDC_CHECK1,BM_GETCHECK,0,0);
					disvbr=MESS(h,IDC_CHECK2,BM_GETCHECK,0,0);
					EndDialog(h,l);
					break;
			}
			break;
		case WM_INITDIALOG:
			MESS(h,IDC_CHECK1,BM_SETCHECK,dmix?BST_CHECKED:BST_UNCHECKED,0);
			MESS(h,IDC_CHECK2,BM_SETCHECK,disvbr?BST_CHECKED:BST_UNCHECKED,0);
			return 1;
	}
	return 0;
}

static void GetMeta(AACSTREAM* stream)
{
	int tc=mp4ff_meta_get_num_items(stream->mp4file);
	if(!tc) return;
	char* item,*value;
	char** tags=(char**)malloc(tc*sizeof(char*));
	int a;
	for(a=0;a<tc;a++) {
		item=NULL;
		value=NULL;
		if(mp4ff_meta_get_by_index(stream->mp4file,a,&item,&value)) {
			tags[a]=(char*)malloc(strlen(item)+strlen(value)+2); // allocate memory for tag
			sprintf(tags[a],"%s=%s",item,value);
		}
	}
	stream->tags=tags;
	stream->tagc=tc;
}

// MP4 file callbacks
uint32_t read_callback(void* user_data,void* buffer,uint32_t length)
{
	AACSTREAM* stream=(AACSTREAM*)user_data;
	DWORD count=length;
	if(!count) return 0;
	while(1) {
		DWORD bytes=xmpfuncs->file.Read(stream->f,buffer,count);
		if(!bytes&&xmpfuncs->file.NetIsActive(stream->f)) {
			if(!stream->buflen) return bytes;
			if(xmpfuncs->file.NetPreBuf(stream->f)) continue;
		}
		if(!bytes) stream->eof=TRUE;
		return bytes;
	}
}

uint32_t seek_callback(void* user_data,uint64_t position)
{
	AACSTREAM* stream=(AACSTREAM*)user_data;
	if(!xmpfuncs->file.Seek(stream->f,position)) return 1;
	stream->eof=FALSE;
	return 0;
}

static long GetAACTrack(mp4ff_t* infile)
{
	/* find AAC track */
	long i;
	char rc;
	for(i=0;i<mp4ff_total_tracks(infile);i++) {
		BYTE* buf=NULL;
		unsigned int buf_size=0;
		mp4AudioSpecificConfig mp4ASC;
		mp4ff_get_decoder_config(infile,i,&buf,&buf_size);
		if(buf) {
			rc=NeAACDecAudioSpecificConfig(buf,buf_size,&mp4ASC);
			free(buf);
			if(rc<0) continue;
			return i;
		}
	}
	/* can't decode this */
	return -1;
}

#define StringComp(str1,str2,len) strncmp(str1,str2,len)

#define ADIF_MAX_SIZE 30 /* Should be enough */
#define ADTS_MAX_SIZE 10 /* Should be enough */

static int sample_rates[]={96000,88200,64000,48000,44100,32000,24000,22050,16000,12000,11025,8000};

static int read_ADIF_header(AACSTREAM *aac)
{
	faadAACInfo* info=&aac->info;
	int bitstream;
	BYTE buffer[ADIF_MAX_SIZE];
	int skip_size = 0;
	int sf_idx;

	/* Get ADIF header data */
	info->headertype = 1;

	if(!xmpfuncs->file.Read(aac->f,buffer,ADIF_MAX_SIZE))
		return -1;

	/* copyright string */
	if(buffer[4] & 0x80)
		skip_size += 9; /* skip 9 bytes */

	bitstream=buffer[4 + skip_size] & 0x10;
	info->bitrate = ((unsigned int)(buffer[4 + skip_size] & 0x0F)<<19)|
		((unsigned int)buffer[5 + skip_size]<<11)|
		((unsigned int)buffer[6 + skip_size]<<3)|
		((unsigned int)buffer[7 + skip_size] & 0xE0);

	if (bitstream==0)
	{
		info->object_type =  ((buffer[6 + skip_size]&0x01)<<1)|((buffer[7 + skip_size]&0x80)>>7);
		sf_idx = (buffer[7 + skip_size]&0x78)>>3;
	} else {
		info->object_type = (buffer[4 + skip_size] & 0x18)>>3;
		sf_idx = ((buffer[4 + skip_size] & 0x07)<<1)|((buffer[5 + skip_size] & 0x80)>>7);
	}
	info->sampling_rate = sample_rates[sf_idx];
	aac->length=8.f*xmpfuncs->file.GetSize(aac->f)/info->bitrate+0.5;

	return 0;
}

static int read_ADTS_header(AACSTREAM *aac,int no_seek_table)
{
	int filetype=xmpfuncs->file.GetType(aac->f);
	faadAACInfo* info=&aac->info;
	/* Get ADTS header data */
	BYTE buffer[ADTS_MAX_SIZE];
	int frames, nextsec=0, t_framelength = 0, frame_length, ID;
	int second = 0, bytes;
	DWORD pos;

	if (filetype>=XMPFILE_TYPE_NETFILE)
               no_seek_table=1; // seek table only available for local files

	info->headertype = 2;

	/* Read all frames to ensure correct time and bitrate */
	for(frames=0;;frames++)
	{
		pos = xmpfuncs->file.Tell(aac->f);

		/* 12 bit SYNCWORD */
		bytes = xmpfuncs->file.Read(aac->f,buffer,ADTS_MAX_SIZE);
		if (bytes!=ADTS_MAX_SIZE) break;
		/* check syncword */
		if ((*(DWORD*)buffer&0xf6ff)!=0xf0ff)
			break;

		if (!frames) {
			/* fixed ADTS header is the same for every frame, so we read it only once */
			/* Syncword found, proceed to read in the fixed ADTS header */
			ID = buffer[1] & 0x08;
			if (!ID)
				info->version = 4;
			else /* MPEG-2 */
				info->version = 2;
			info->object_type = (buffer[2]&0xC0)>>6;
			info->sampling_rate = sample_rates[(buffer[2]&0x3C)>>2];
			info->channels = ((buffer[2]&0x01)<<2)|((buffer[3]&0xC0)>>6);
		}

		frame_length = ((((unsigned int)buffer[3] & 0x3)) << 11)
			| (((unsigned int)buffer[4]) << 3) | (buffer[5] >> 5);
		t_framelength += frame_length;

		if (!no_seek_table && frames==nextsec) {
			aac->st = (unsigned long *) realloc(aac->st, (second + 1) * sizeof(unsigned long));
			aac->st[second] = pos;
			second++;
			nextsec=second*info->sampling_rate/1024; // start frame of next second
		}
		{
			BYTE tmp[8192]; // 8192=max possible frame size
			frame_length-=ADTS_MAX_SIZE;
			if (filetype>=XMPFILE_TYPE_NETFILE && xmpfuncs->file.NetAvailable(aac->f)<frame_length) {
				frames++;
				break;
			}
			if (xmpfuncs->file.Read(aac->f,tmp,frame_length)!=frame_length)
				break;
		}
	}

	if (frames<2) return -1; // probably not an AAC file!

	if (aac->st) aac->seekable=1;

	info->bitrate = (8.f/1024)*t_framelength*info->sampling_rate/frames;

	if (filetype<XMPFILE_TYPE_NETFILE)
		aac->length=1024.f*frames/info->sampling_rate;
	else if (filetype==XMPFILE_TYPE_NETFILE)
		aac->length=8.f*xmpfuncs->file.GetSize(aac->f)/info->bitrate;

	return frames;
}

static int get_AAC_format(AACSTREAM *aac, int no_seek_table)
{
	XMPFILE file=aac->f;
	faadAACInfo* info=&aac->info;

	memset(info, 0, sizeof(faadAACInfo));

	xmpfuncs->file.Seek(file,0);
	{
		BYTE buf[500];
		int a,b=xmpfuncs->file.Read(file,buf,500);
		for (a=0;a<b-4;a++) {
			if (*(DWORD*)(buf+a)==MAKEFOURCC('A','D','I','F')) { // "ADIF" signature
				aac->initbytes=a;
				xmpfuncs->file.Seek(file,a);
				return read_ADIF_header(aac);
			}
			if (((*(DWORD*)(buf+a)&0xf6ff)==0xf0ff && ((buf[a+2]&0x3c)>>2)<12)) { // ADTS header syncword
				aac->initbytes=a;
				xmpfuncs->file.Seek(file,a);
				return read_ADTS_header(aac,no_seek_table);
			}
		}
	}
	return -1;
}

static void FreeAAC(AACSTREAM* aac)
{
	if(aac->mode) { // MP4
		NeAACDecClose(aac->decoder);
		if(aac->tags) {
			int a;
			for(a=0;a<aac->tagc;a++) free(aac->tags[a]);
			free(aac->tags);
		}
		if(aac->mp4file) mp4ff_close(aac->mp4file);
	} else { // AAC
		free(aac->st);
		NeAACDecClose(aac->decoder);
	}
	free(aac->fbuf);
	free(aac);
}

// check if a file is playable by this plugin
// more thorough checks can be saved for the GetFileInfo and Open functions
// length = size of "buf" memory block
static BOOL WINAPI AAC_CheckFile(char* filename,BYTE* buf,DWORD length)
{
	char* ext=strrchr(filename,'.');
	if(ext) {
		if(!stricmp(ext+1,"MP4")||!stricmp(ext+1,"M4A")||!stricmp(ext+1,"M4B")) {
			if(*(DWORD*)(buf+4)==MAKEFOURCC('f','t','y','p')) return 1;
			return 0;
		} else if(!stricmp(ext+1,"TXT")||!stricmp(ext+1,"JPG")||!stricmp(ext+1,"GIF")||!stricmp(ext+1,"XML")) return 0; // there are tons of more exts to add
	}
	int a;
	for (a=0;a<(int)min(length,500)-4;a++) // limited header scanning to 500 (instead of 1000)
		if (*(DWORD*)(buf+a)==MAKEFOURCC('A','D','I','F') // "ADIF" signature
			|| ((*(DWORD*)(buf+a)&0xf6ff)==0xf0ff && ((buf[a+2]&0x3c)>>2)<12)) // ADTS header syncword
			return 1;
	return 0;
}

static AACSTREAM* InitAAC(XMPFILE file,int st)
{
	AACSTREAM* aac=(AACSTREAM*)calloc(1,sizeof(AACSTREAM));
	aac->f=file;
	aac->mode=0;
	if((aac->frames=get_AAC_format(aac,st))==-1) {
		FreeAAC(aac);
		return NULL;
	}
	if (st) return aac; // skip decoder initialization if just getting info

	if(!(aac->decoder=NeAACDecOpen())) {
		FreeAAC(aac);
		return NULL;
	}
	NeAACDecConfigurationPtr config=NeAACDecGetCurrentConfiguration(aac->decoder);
	config->downMatrix=dmix;
	config->outputFormat=FAAD_FMT_FLOAT;
	NeAACDecSetConfiguration(aac->decoder,config);

	BYTE buffer[FAAD_MIN_STREAMSIZE*6];
	int inb,outb;
	xmpfuncs->file.Seek(aac->f,aac->initbytes);
	inb=xmpfuncs->file.Read(aac->f,buffer,FAAD_MIN_STREAMSIZE*6);
	if((outb=NeAACDecInit(aac->decoder,buffer,inb,&aac->sr,&aac->nch))<0) {
		FreeAAC(aac);
		return NULL;
	}
	if(dmix&&(aac->nch==5||aac->nch==6)) aac->nch=2;
	aac->initbytes+=outb;
	xmpfuncs->file.Seek(aac->f,aac->initbytes);
	return aac;
}

static AACSTREAM* InitMP4(XMPFILE file)
{
	AACSTREAM* aac=(AACSTREAM*)calloc(1,sizeof(AACSTREAM));
	aac->f=file;
	aac->mode=1;
	xmpfuncs->file.Seek(aac->f,0);
	BYTE* buffer;
	unsigned int bs; // buffer size
	aac->mp4cb.read=read_callback;
	aac->mp4cb.seek=seek_callback;
	aac->mp4cb.user_data=aac;
	if(!(aac->mp4file=mp4ff_open_read(&aac->mp4cb))) {
		FreeAAC(aac);
		return NULL;
	}
	if((aac->mp4track=GetAACTrack(aac->mp4file))<0) {
		FreeAAC(aac);
		return NULL;
	}
	if(mp4ff_get_decoder_config(aac->mp4file,aac->mp4track,&buffer,&bs)) {
		FreeAAC(aac);
		return NULL;
	}
	if(!(aac->decoder=NeAACDecOpen())
		|| NeAACDecInit2(aac->decoder,buffer,bs,&aac->sr,&aac->nch)<0) {
		free(buffer);
		FreeAAC(aac);
		return NULL;
	}
	if(!(aac->timescale=mp4ff_time_scale(aac->mp4file,aac->mp4track))) {
		free(buffer);
		FreeAAC(aac);
		return NULL;
	}
	double timescale_div=1.0/(double)aac->timescale;
	{
		int64_t duration=mp4ff_get_track_duration_use_offsets(aac->mp4file,aac->mp4track);
		if(duration==-1)
			aac->length=0;
		else
			aac->length=(float)duration*timescale_div;
	}
	NeAACDecConfigurationPtr config;
	config=NeAACDecGetCurrentConfiguration(aac->decoder);
	config->downMatrix=dmix;
	config->outputFormat=FAAD_FMT_FLOAT;
	NeAACDecSetConfiguration(aac->decoder,config);
	if(dmix&&(aac->nch==5||aac->nch==6)) aac->nch=2;
	aac->ns=mp4ff_num_samples(aac->mp4file,aac->mp4track);
	aac->sampleid=0;
	aac->seekable=1;
	{
		GetMeta(aac);
		free(buffer);
	}
	return aac;
}

static AACSTREAM* Init(char* filename,XMPFILE file,int st)
{
	char* ext=strrchr(filename,'.');
	if(ext)
		if(!stricmp(ext+1,"MP4")||!stricmp(ext+1,"M4A")||!stricmp(ext+1,"M4B")) return InitMP4(file);
	return InitAAC(file,st);
}

static void WINAPI AAC_About(HWND win)
{
	MessageBox(win,"AAC/MP4 Plugin for XMPlay 3.2\nPortions Copyright  2003, M. Bakker, Nero AG, http://www.nero.com\nCopyright  2005, MaresWEB","About",MB_OK);
}

static void WINAPI AAC_Config(HWND win)
{
	DialogBox(inst,MAKEINTRESOURCE(IDD_DIALOG1),win,&DialogProc);
}

// get iTunes/MP4 metadata from an AACSTREAM
// ID3v1/v2 & APEv2 tags may already be in the "tags" array (don't free them - XMPlay will if replaced)
static void GetTags(AACSTREAM* aac, char *tags[7])
{
	static char* tagname[7]={"title=","artist=","album=","date=","track=","genre=","comment="};
	static int taglen[7]={6,7,6,5,6,6,8};
	if(aac->tags) {
		char* temp[7]={0}; // use a temporary array, in case of duplicate tags
		int a,b;
		for(a=0;a<aac->tagc;a++) {
			char* tag=aac->tags[a];
			for(b=0;b<7;b++) {
				if(!strnicmp(tag,tagname[b],taglen[b])) { // found a tag, copy it...
					if(temp[b]) { // already an entry for this tag, add to it
						int l=strlen(temp[b]);
						temp[b]=(char*)realloc(temp[b],l+strlen(tag+taglen[b])+2);
						sprintf(temp[b]+l,"/%s",tag+taglen[b]);
					} else
						temp[b]=strdup(tag+taglen[b]);
					break;
				}
			}
		}
		for(a=0;a<7;a++)
			if(temp[a]) tags[a]=temp[a];
	}
}

// get file info (length & tags);
static BOOL WINAPI AAC_GetFileInfo(char* filename,XMPFILE file,float* length,char* tags[7])
{
	AACSTREAM* aac=Init(filename,file,1);
	if(!aac) return 0; // failed
	if(length) *length=aac->length;
	if(aac->mode) GetTags(aac,tags);
	FreeAAC(aac);
	return 1; // success
}

// open a file for playback
// return:  0=failed, 1=success, 2=success and XMPlay can close the file
static DWORD WINAPI AAC_Open(char* filename,XMPFILE file)
{
	aac=Init(filename,file,0);
	if(!aac) return 0; // failed
	aac->fbuf=(BYTE*)malloc(FAAD_MIN_STREAMSIZE*6);
	{
		xmpfuncs->SetLength(aac->length,aac->seekable); // set length
		if (aac->info.bitrate) aac->abrate=aac->brate=aac->info.bitrate/8;
		else aac->abrate=aac->brate=xmpfuncs->file.GetSize(file)/aac->length;
		if(xmpfuncs->file.GetType(file)>XMPFILE_TYPE_FILE)
			xmpfuncs->file.NetSetRate(file,aac->abrate);
	}
	if(aac->mode && aac->tags) { // check for replaygain
		int a;
		char* tag;
		for(a=0;a<aac->tagc;a++) {
			tag=aac->tags[a];
			if(!strnicmp(tag,"replaygain_track_gain=",22))
				xmpfuncs->SetGain(0,atof(tag+22)); // set "track" gain
			else if(!strnicmp(tag,"replaygain_album_gain=",22))
				xmpfuncs->SetGain(1,atof(tag+22)); // set "album" gain
		}
	}
	return 1;
}

// close the file
static void WINAPI AAC_Close()
{
	FreeAAC(aac);
	aac=0;
}

// get the tags
// return TRUE to delay the title update when there are no tags (use UpdateTitle to request update when ready)
static BOOL WINAPI AAC_GetTags(char* tags[7])
{
	if(aac->mode) GetTags(aac,tags);
	return 0;
}

// return the sample format
static DWORD WINAPI AAC_GetFormat(DWORD* chan,DWORD* res)
{
	if(res) *res=0; // bytes per sample (0=not a fixed resolution, eg. lossy encoding)
	if(chan) *chan=aac->nch;
	return aac->sr;
}

// get the main panel info text
static void WINAPI AAC_GetInfoText(char* format,char* length)
{
	if(format) sprintf(format,"%s - %dkb/s - %dhz",aac->mode?"MP4":"AAC",(disvbr?aac->brate:aac->abrate)/125,aac->sr); // format details
// "length" will already contain the time, but more info can be added to the end, eg...
	if((length&&!aac->mode&&aac->info.headertype==ADTS&&xmpfuncs->file.GetType(aac->f)<XMPFILE_TYPE_NETFILE)) sprintf(strchr(length,0)," - %d frames",aac->frames);
}

// get text for "General" info window
// separate headings and values with a tab (\t), end each line with a carriage-return (\r)
static void WINAPI AAC_GetGeneralInfo(char* buf)
{
	buf+=sprintf(buf,"Format\tAdvanced Audio Coding (%s)\r",aac->mode?"MP4":"AAC");
	if(!aac->mode)
		sprintf(buf,"Header type\t%s\rSample rate\t%d hz\rBit rate\t%d kbps\rChannels\t%d\r",
		htypes[aac->info.headertype],aac->sr,aac->abrate/125,aac->nch);
	else
		sprintf(buf,"Sample rate\t%d hz\rBit rate\t%d kbps\rChannels\t%d\r",
		aac->sr,aac->abrate/125,aac->nch);
}

// get text for "Message" info window
// separate tag names and values with a tab (\t), and end each line with a carriage-return (\r)
static void WINAPI AAC_GetMessage(char* buf)
{
	if(aac->tags) {
		int a;
		for(a=0;a<aac->tagc;a++) {
			char* p;
			sprintf(buf,"%s\r",aac->tags[a]);
			if(p=strchr(buf,'=')) {
				*p='\t';
				while((p=strchr(p,'\r'))&&*++p) { // need to add tabs at start of each new line
					if(*p!='\n') memmove(p+1,p,strlen(p)+1);
					*p='\t';
				}
			}
			buf=strchr(buf,0);
		}
	}
}

// get some sample data, always floating-point
// count=number of floats to write (not bytes or samples)
// return number of floats written. if it's less than requested, playback is ended...
// so wait for more if there is more to come (use CheckCancel function to check if user wants to cancel)
static DWORD WINAPI AAC_Process(float* buffer,DWORD count)
{
	DWORD done=0,c;
	while(done<count) {
		if(aac->bufp==aac->buflen) {
			NeAACDecFrameInfo info;
			if(!aac->mode)
				aac->fbuflen+=xmpfuncs->file.Read(aac->f,aac->fbuf+aac->fbuflen,FAAD_MIN_STREAMSIZE*6-aac->fbuflen);
			else {
				if(aac->sampleid>=aac->ns) break;
				aac->fbuflen+=mp4ff_read_sample_v2(aac->mp4file,aac->mp4track,aac->sampleid++,aac->fbuf+aac->fbuflen);
			}
			aac->buf=(float*)NeAACDecDecode(aac->decoder,&info,aac->fbuf,aac->fbuflen);
			
			if(!dmix&&info.channels==6&&info.num_lfe_channels) {
				//channel order for 5.1: L/R/C/LF/BL/BR
				float r1,r2,r3,r4,r5,r6;
				unsigned int i;
				for(i=0;i<info.samples;i+=info.channels/*6*/) {
					r1=aac->buf[i];
					r2=aac->buf[i+1];
					r3=aac->buf[i+2];
					r4=aac->buf[i+3];
					r5=aac->buf[i+4];
					r6=aac->buf[i+5];
					aac->buf[i]=r2;
					aac->buf[i+1]=r3;
					aac->buf[i+2]=r1;
					aac->buf[i+3]=r6;
					aac->buf[i+4]=r4;
					aac->buf[i+5]=r5;
				}
			}

			aac->fbuflen-=info.bytesconsumed;
			memmove(aac->fbuf,aac->fbuf+info.bytesconsumed,aac->fbuflen); // shift remaining data to start of buffer
			if(!aac->buf) {
				if (xmpfuncs->file.NetIsActive(aac->f)) { // still connected...
					if (xmpfuncs->file.NetPreBuf(aac->f)) // pre-buffer some more
						continue;
				}
				break;
			}
			if (!info.samples) continue;
			aac->buflen=info.samples;
			aac->bufp=0;

			if(disvbr) {
				aac->seq_samples+=info.samples;
				aac->seq_bytes+=info.bytesconsumed;
				float secs=(float)aac->seq_samples/info.channels/info.samplerate;
				if (secs>=1) {
					aac->brate=aac->seq_bytes/secs;
					aac->seq_bytes=aac->seq_samples=0;
					xmpfuncs->RefreshInfo(XMPINFO_MAIN);
				}
			}
		}
		c=min(count-done,aac->buflen-aac->bufp);
		memcpy(buffer+done,aac->buf+aac->bufp,c*sizeof(float));
		aac->bufp+=c;
		done+=c;
	}
	return done;
}

// Get the seeking granularity in seconds
static double WINAPI AAC_GetGranularity()
{
	return aac->mode?0.1f:1; // seek in units of a 1/10th of a second (1 sec for AAC)
}

// Seek to a position (in granularity units)
// return the new position in seconds
static double WINAPI AAC_SetPos(DWORD pos)
{
	aac->fbuflen=aac->bufp=aac->buflen=0;
	if(!aac->mode) {
		int time=pos;
		if(!xmpfuncs->file.Seek(aac->f,aac->st[time])) {
			if(!pos) return -1; // restart failed
			while(time>0) { // go back 3 seconds and try again
				time-=3;
				if(time<0) time=0;
				if(xmpfuncs->file.Seek(aac->f,aac->st[time]))
					break;
			}
		}
		return time;
	} else {
		double time=pos*AAC_GetGranularity(); // convert "pos" into seconds
		int64_t offset=(int64_t)(time*aac->timescale+0.5);
		aac->sampleid=mp4ff_find_sample_use_offsets(aac->mp4file,aac->mp4track,offset);
		return time;
	}
}

// plugin function table
static XMPIN xmpin={
	XMPIN_FLAG_CANSTREAM,
	"Advanced Audio Coding",
	"Advanced Audio Coding\0aac/mp4/m4a/m4b",
	AAC_About,
	AAC_Config,
	AAC_CheckFile,
	AAC_GetFileInfo,
	AAC_Open,
	AAC_Close,
	AAC_GetFormat,
	NULL, // SetFormat doesn't apply (XMPlay gets the format, not sets it)
	AAC_GetTags,
	AAC_GetInfoText,
	AAC_GetGeneralInfo,
	AAC_GetMessage,
	AAC_SetPos,
	AAC_GetGranularity,
	NULL, // GetBuffering only applies when using your own file routines
	AAC_Process,
	NULL, // WriteFile, see GetBuffering
};

// get the plugin's XMPIN interface
XMPIN* WINAPI XMPIN_GetInterface(DWORD face,XMPIN_FUNCS* funcs)
{
	if(face!=XMPIN_FACE) return NULL;
	xmpfuncs=funcs;
	return &xmpin;
}

BOOL WINAPI DllMain(HINSTANCE hDLL, DWORD reason, LPVOID reserved)
{
	switch(reason) {
	case DLL_PROCESS_ATTACH:
		{
			inst=hDLL;
			DisableThreadLibraryCalls(hDLL);
			//isWinNT=(int)GetVersion()>=0; // useful to know when doing your own file opening

			HKEY reg;
			DWORD s;
			if(!RegCreateKeyEx(HKEY_CURRENT_USER,"SOFTWARE\\Un4seen Developments\\XMPlay2\\AAC",0,NULL,0,KEY_ALL_ACCESS,NULL,&reg,&s)) {
				s=sizeof(dmix);
				RegQueryValueEx(reg,"Downmatrix",0,0,(BYTE*)&dmix,&s);
				s=sizeof(disvbr);
				RegQueryValueEx(reg,"DisplayVBR",0,0,(BYTE*)&disvbr,&s);
				RegCloseKey(reg);
			}
		}
		break;
	case DLL_PROCESS_DETACH:
		{
			HKEY reg;
			DWORD s;
			if(!RegCreateKeyEx(HKEY_CURRENT_USER,"SOFTWARE\\Un4seen Developments\\XMPlay2\\AAC",0,NULL,0,KEY_ALL_ACCESS,NULL,&reg,&s)) {
				RegSetValueEx(reg,"Downmatrix",0,REG_DWORD,(BYTE*)&dmix,4);
				RegSetValueEx(reg,"DisplayVBR",0,REG_DWORD,(BYTE*)&disvbr,4);
				RegCloseKey(reg);
			}
		}
		break;
	}
	return TRUE;
}
